import 'dart:collection';

import 'package:meta/meta.dart';


/// `DelegateMapMixin` can be used as a mixin to implement a delegating map. It
/// delegates all methods to [delegate].
mixin DelegateMapMixin<K, V> on MapMixin<K, V> {

  @override
  V operator[](Object key) => delegate[key];

  @override
  void operator[]=(K key, V value) => delegate[key] = value;

  @override
  void clear() => delegate.clear();

  @override
  Iterable<K> get keys => delegate.keys;

  @override
  V remove(Object key) => delegate.remove(key);

  /// The underlying map to which all methods are delegated.
  @visibleForOverriding
  @protected Map<K, V> get delegate;

}


/// A map that delegates all methods to an underlying map.
class DelegateMap<K, V> with MapMixin<K, V>, DelegateMapMixin<K, V> {

  final Map<K, V> _map;

  /// Creates a `DelegateMap` that delegates all methods to the given [_map].
  DelegateMap(this._map);


  @override
  @visibleForTesting
  @protected Map<K, V> get delegate => _map;

}