import 'package:sane/core.dart';

import 'delegate_map.dart';
import 'observable_modification.dart';


/// A map that emits [Modification] events when it is modified. All methods are delegated
/// to an underlying map.
class ObservableMap<K, V> extends DelegateMap<K, V> with Listenable<Modification<K, V>> {

  /// Creates an `ObservableMap` that delegates all methods to [map].
  /// If [map] is not given, a `LinkedHashMap` is used.
  ObservableMap([Map<K, V> map]) : super(map ?? {});


  /// Associates the [key] with the given [value]; always emits a `Modification.add` event.
  @override
  void operator[]=(K key, V value) {
    delegate[key] = value;
    source.add(Modification.add(key, value));
  }

  /// Removes [key] and its associated `value`, if present, from the map; always
  /// emits a `Modification.remove` event.
  ///
  /// Returns the value associated with [key] before it was removed. Returns `null` if [key] was not in the map.
  //
  // Note that values can be `null` and a returned `null` value doesn't always mean that the key was absent.
  @override
  V remove(Object key) {
    final value = delegate.remove(key);
    source.add(Modification.remove(key, value));
    return value;
  }

  /// Removes all entries from this map; always emits a `Modification.clear` event.
  @override
  void clear() {
    delegate.clear();
    source.add(const Modification.clear());
  }

}

