import 'package:sane/core.dart';

import 'delegate_set.dart';
import 'observable_modification.dart';


/// A set that emits [Modification] events when it is modified. Emitted events do
/// not contain a key. All methods are delegated to an underlying set.
class ObservableSet<T> extends DelegateSet<T> with Listenable<Modification<void, T>> {

  /// Creates an `ObservableSet` that delegates all methods to [set].
  /// If [set] is not given, a `LinkedHashSet` is used.
  ObservableSet([Set<T> set]) : super(set ?? <T>{});


  /// Adds [value] to the set; always emits a `Modification.add` event.
  ///
  /// Returns `true` if [value] (or an equal value) was not yet in the set. Otherwise
  /// returns `false` and the set is not changed.
  @override
  bool add(T value) {
    final contains = delegate.add(value);
    source.add(Modification.add(null, value));
    return contains;
  }

  /// Removes [value] from the set; always emit a `Modification.remove` event.
  ///
  /// Returns `true` if [value] was in the set. Returns `false` otherwise.
  @override
  bool remove(Object value) {
    final removed = delegate.remove(value);
    source.add(Modification.remove(null, value));
    return removed;
  }

  /// Removes all elements in the set; always emit a `Modification.clear` event.
  @override
  void clear() {
    delegate.clear();
    source.add(const Modification.clear());
  }

}