import 'package:meta/meta.dart';

import 'package:sane/core.dart';


/// A function that produces a result using the given argument.
typedef Transform<T, R> = R Function(T type);


/// An immutable key-value pair.
@immutable class Pair<K, V> implements MapEntry<K, V> {

  /// The key.
  @override
  final K key;
  /// The value.
  @override
  final V value;

  /// Creates a `Pair` with the given [key] and [value].
  @literal const Pair(this.key, this.value);


  /// Copies `this` with the given replacement [key] and/or [value].
  Pair<K, V> replace({K key, V value}) => Pair(key ?? this.key, value ?? this.value);

  /// Returns a `Pair` derived using the given [keyer] and [valuer] on [key] and [value] respectively.
  Pair<K2, V2> map<K2, V2>(Transform<K, K2> keyer, Transform<V, V2> valuer) => Pair(keyer(key), valuer(value));

  /// Returns a value using the given [reduction] on [key] and [value].
  T reduce<T>(T Function(K, V) reduction) => reduction(key, value);

  /// Returns a list that contains [key] and [value].
  List<dynamic> toList({bool immutable = false}) => immutable ? [key, value] : List.unmodifiable([key, value]);


  @override
  bool operator == (Object other) => other is Pair && key == other.key && value == other.value;

  @override
  int get hashCode => hash([key, value]);

  @override
  String toString() => 'Pair{$key: $value}';

}


/// An immutable triple.
@immutable class Triple<L, M, R> {

  /// The left value.
  final L left;
  /// The middle value.
  final M middle;
  /// The right value.
  final R right;

  /// Creates a `Triple` with the given [left], [middle] and [right] values.
  @literal const Triple(this.left, this.middle, this.right);

  /// Copies `this` with the given replacement [left], [middle] and/or [right] values.
  Triple<L, M, R> replace({L left, M middle, R right}) => Triple(left ?? this.left, middle ?? this.middle, right ?? this.right);

  /// Returns a `Triple` derived using the given [lefter], [middler] and [righter] on
  /// [left], [middle] and [right] respectively.
  Triple<L2, M2, R2> map<L2, M2, R2>(Transform<L, L2> lefter, Transform<M, M2> middler, Transform<R, R2> righter) => Triple(lefter(left), middler(middle), righter(right));

  /// Returns a value using the given [reduction] on [left], [middle] and [right].
  T reduce<T>(T Function(L, M, R) reduction) => reduction(left, middle, right);


  /// Returns a list that contains [left], [middle] and [right].
  List<dynamic> toList({bool immutable = false}) => immutable ? [left, middle, right] : List.unmodifiable([left, middle, right]);


  @override
  bool operator == (Object other) => other is Triple && left == other.left && middle == other.middle && right == other.right;

  @override
  int get hashCode => hash([left, middle, right]);

  @override
  String toString() => 'Triple{$left, $middle, $right}';

}


/// An immutable tuple that contains 4 values.
@immutable class Quad<T1, T2, T3, T4> {

  /// The first value.
  final T1 first;
  /// The second value.
  final T2 second;
  /// The third value.
  final T3 third;
  /// The fourth value.
  final T4 fourth;

  //// Creates a `Quad` with the given [first], [second], [third] and [fourth] values.
  @literal const Quad(this.first, this.second, this.third, this.fourth);


  /// Copies `this` with the given replacement [first], [second], [third] and/or [fourth] values.
  Quad<T1, T2, T3, T4> replace({T1 first, T2 second, T3 third, T4 fourth}) => Quad(first ?? this.first, second ?? this.second, third ?? this.third, fourth ?? fourth);

  /// Returns a `Quad` derived using the given [first], [second], [third] and [fourth] on
  /// [first], [second], [third] and [fourth] values respectively.
  Quad<R1, R2, R3, R4> map<R1, R2, R3, R4>(Transform<T1, R1> first, Transform<T2, R2> second, Transform<T3, R3> third, Transform<T4, R4> fourth)
    => Quad(first(this.first), second(this.second), third(this.third), fourth(this.fourth));

  /// Returns a value using the given [reduction] on [first], [second], [third] and [fourth].
  T reduce<T>(T Function(T1, T2, T3, T4) reduction) => reduction(first, second, third, fourth);


  /// Creates a list that contains [first], [second], [third] and [fourth].
  List<dynamic> toList({bool immutable = false}) => immutable ? [first, second, third, fourth] : List.from([first, second, third, fourth], growable: true);


  @override
  bool operator == (Object other) => other is Quad && first == other.first && second == other.second && third == other.third && fourth == other.fourth;

  @override
  int get hashCode => hash([first, second, third, fourth]);

  @override
  String toString() => 'Quad{$first, $second, $third, $fourth}';

}