import 'dart:collection';

import 'package:meta/meta.dart';
import 'package:sane/core.dart';


/// A map that associates a pseudo weak key to a value. An entry is automatically
/// removed once it is unreachable via [handle]. All methods are delegated
/// to an underlying map.
class WeakMap<K, V> extends Weak<K> with Janitor<K>, MapMixin<K, V> {

  /// The underlying map to which all methods are delegated.
  @visibleForTesting
  final Map<K, V> delegate;

  /// Creates a `WeakMap` with the given [handle] and underlying [map]. If [map] is not given,
  /// a `LinkedHashMap` is used.
  WeakMap(Handle<K> handle, [Map<K, V> map]): delegate = map ?? {}, super(handle);


  @override
  void operator[]=(K key, V value) {
    if (handle.contains(key)) {
      delegate[key] = value;
      modifications--;
    }
  }

  @override
  @nullable V remove(Object key) {
    final value = delegate.remove(key);
    if (value != null) {
      modifications--;
    }

    return value;
  }

  @override
  void clear() {
    modifications = handle.modifications;
    delegate.clear();
  }


  @override
  @nullable V operator[](Object key) {
    clean();
    return delegate[key];
  }


  @override
  Iterable<K> get keys {
    clean();
    return delegate.keys;
  }


  @override
  @visibleForTesting
  void cleanup() => delegate.removeWhere((key, value) => !handle.contains(key));

}