import 'dart:collection';

import 'package:meta/meta.dart';
import 'package:sane/core.dart';


/// A set that contains pseudo weak values. A value is automatically removed once
/// it is unreachable via [handle]. All methods are delegated to an underlying set.
class WeakSet<T> extends Weak<T> with Janitor<T>, SetMixin<T> {

  /// The underlying set to which all methods are delegated.
  @visibleForTesting
  final Set<T> set;

  /// Creates a `WeakSet` with the given [handle] and underlying [set]. If [set] is not given,
  /// a `LinkedHashSet` is used.
  WeakSet(Handle<T> handle, [Set<T> set]): set = set ?? {}, super(handle);


  @override
  bool add(T value) {
    final added = handle.contains(value);
    if (added) {
      set.add(value);
      modifications--;
    }

    return added;
  }

  @override
  bool remove(Object value) {
    final removed = set.remove(value);
    if (removed) {
      modifications--;
    }

    return removed;
  }

  @override
  void clear() {
    modifications = handle.modifications;
    set.clear();
  }


  @override
  @nullable T lookup(Object element) => contains(element) ? element : null;

  @override
  bool contains(Object element) {
    clean();
    return set.contains(element);
  }


  @override
  Iterator<T> get iterator {
    clean();
    return set.iterator;
  }

  @override
  int get length {
    clean();
    return set.length;
  }


  @override
  WeakSet<T> toSet() {
    clean();
    return WeakSet(handle, set);
  }


  @override
  @visibleForTesting
  void cleanup() => set.removeWhere((element) => !handle.contains(element));

}