import 'package:meta/meta.dart';


/// A handle used by pseudo weak types to determine the reachability of a value.
///
/// When a value is no longer reachable, [mods] must be incremented and [contains]
/// must return `false`.
mixin Handle<T> {

  /// The modification counter that must be increased each time a value is no longer
  /// reachable.
  @visibleForTesting
  @protected int mods = 0;


  /// Returns `true` if the given `referent` is reachable and `false` otherwise.
  bool contains(T referent);

  /// The modification counter.
  int get modifications => mods;

}


/// A handle that always returns the value that it was initialized with.
@immutable class ConstantHandle<T> implements Handle<T> {

  /// The value which is always returned by [contains].
  final bool value;

  /// Creates a `ConstantHandle` which always returns the given [value].
  const ConstantHandle({@required this.value});

  /// Determines if the given `referent` is reachable; always returns [value].
  @override
  bool contains(T referent) => value;

  /// The modification counter; always `0`.
  @override
  int get modifications => 0;


  /// The modification counter; always `0`.
  @override
  @visibleForTesting
  int get mods => 0;

  /// Does nothing.
  @override
  @visibleForTesting
  set mods(int mods) {}

}


/// A handle that determines the reachability of a value by checking for its existence
/// in a backing `set`.
abstract class SetHandle<T> with Handle<T> {

  /// The backing set.
  @protected final Set<T> set;

  /// Creates a `SetHandle` with the given backing [set].
  SetHandle(this.set);


  @override
  bool contains(T referent) => set.contains(referent);

}


/// A handle that determines the reachability of a value by checking for its existence
/// in a backing `map`.
abstract class MapHandle<K, V> with Handle<K> {

  /// The backing map.
  @protected final Map<K, V> map;

  /// Creates a `MapHandle` with the given backing [map].
  MapHandle(this.map);


  @override
  bool contains(K referent) => map.containsKey(referent);

}