import 'package:sane/core.dart';


/// Utilities for formatting numbers.
extension FormattedNumber<T extends num> on T {

  /// Transforms `this` into a `String` and pads it on the left if it is shorter
  /// than the given [width].
  String padLeft(int width, [String padding = '0']) => toString().padLeft(width, padding);

  /// Transforms `this` into a `String` and pads it on the right if it is shorter
  /// than the given [width].
  String padRight(int width, [String padding = '0']) => toString().padRight(width, padding);

}

/// Utilities for determining the range of numbers.
extension RangedNumber<T extends num> on T {

  /// Returns `true` if `this` is between the given [min] (inclusive) and [max] (exclusive).
  bool between(num min, num max) {
    assert(min <= max, 'Invalid bounds: $min and $max, min cannot be greater than max');
    return min <= this && this < max;
  }

  /// Returns `true` if this number is outside the given range of [min] (exclusive) and
  /// [max] (exclusive).
  bool outside(num min, num max) {
    assert(min <= max, 'Invalid bounds: $min and $max, min cannot be greater than max');
    return this < min || this > max;
  }

}

/// Utilities for rounding numbers.
extension RoundableNumber<T extends num> on T {

  /// Rounds `this` to the nearest given [value].
  T roundTo(@nonZero num value) {
    assert(value != 0, 'Number cannot be rounded about 0');
    return value == 1 ? this : (this / value).round() * value;
  }

  /// Ceils `this` to the nearest given [value].
  T ceilTo(@nonZero num value) {
    assert(value != 0, 'Number cannot be ceiled about 0');
    return value == 1 ? this : (this / value).ceil() * value;
  }

  /// Floors `this` to the nearest given [value].
  T floorTo(@nonZero num value) {
    assert(value != 0, 'Number cannot be floored about 0');
    return value == 1 ? this : (this / value).floor() * value;
  }

}

/// Produces a hash using the given [values].
int hash(Iterable<dynamic> values) {
  var hash = 17;
  for (final value in values) {
    hash = 37 * hash + value.hashCode;
  }

  return hash;
}