import 'package:meta/meta.dart';

import 'annotations.dart';
import 'handles.dart';


/// A skeleton implementation of a pseudo weak type. Implementations should use
/// [handle] to determine the reachability of values. [modifications] should also
/// be decreased each time the weak type is modified.
abstract class Weak<T> {

  /// The `handle` used to determine the reachability of values.
  @visibleForTesting
  @protected final Handle<T> handle;
  /// The modification counter that should be decreased each time the weak type is modified.
  @visibleForTesting
  @protected int modifications;

  /// Creates a `Weak` with the given [handle] and [modifications] initialized using
  /// `handle.modifications`.
  Weak(this.handle): modifications = handle.modifications;

}


/// A pseudo weak type that does additional housekeeping.
mixin Janitor<T> on Weak<T> {

  /// Performs housekeeping if [modifications] differs from the the modification counter
  /// of [handle]. Delegates housekeeping to [cleanup].
  @visibleForTesting
  @protected void clean() {
    if (modifications != handle.modifications) {
      cleanup();
      modifications = handle.modifications;
    }
  }

  /// Does nothing.
  @visibleForTesting
  @protected void cleanup() {}

}


/// A pseudo weak reference which value is automatically removed once it is unreachable
/// via [handle].
class WeakReference<T> extends Weak<T> {

  T _value;

  /// Creates a `WeakReference` with the given [handle].
  WeakReference(Handle<T> handle, [this._value]) : super(handle);


  /// Returns `true` if the value is still reachable via [handle] and `false` otherwise.
  bool get exists {
    if (_value == null) {
      return false;
    }

    if (modifications == handle.modifications) {
      return true;
    }

    final contains = handle.contains(_value);
    if (contains) {
      modifications = handle.modifications;

    } else {
      _value = null;
    }

    return contains;
  }

  /// Returns the value if reachable via [handle] and `null` otherwise.
  @nullable T get value => (this..exists)._value;

}
