import 'package:meta/meta.dart';

import 'package:sane/core.dart';

import 'period.dart';


/// Utilities for conversion between the different date-time representations.
extension DateTimeConversion on DateTime {

  /// Returns a `GlobalDateTime` that represents `this`.
  GlobalDateTime toGlobal() => GlobalDateTime._of(isUtc ? this : toUtc());

  /// Returns a `LocalDateTime` that represents `this`.
  LocalDateTime toLocalized() => LocalDateTime._of(isUtc ? toLocal(): this);


  /// Returns only the date.
  DateTime get date => isUtc ? GlobalDateTime(year, month, day) : LocalDateTime(year, month, day);

}


/// Utilities for rounding `DateTime`s.
extension RoundableDateTime on DateTime {

  /// Rounds `this` to the nearest given `period`.
  DateTime round(Period period) => _from(
    year.roundTo(period.year),
    month.roundTo(period.month),
    day.roundTo(period.day),
    hour.roundTo(period.hour),
    minute.roundTo(period.minute),
    second.roundTo(period.second),
    millisecond.roundTo(period.millisecond),
    microsecond.roundTo(period.microsecond),
  );

  /// Ceils `this` to the nearest given `period`.
  DateTime ceil(Period period) => _from(
    year.ceilTo(period.year),
    month.ceilTo(period.month),
    day.ceilTo(period.day),
    hour.ceilTo(period.hour),
    minute.ceilTo(period.minute),
    second.ceilTo(period.second),
    millisecond.ceilTo(period.millisecond),
    microsecond.ceilTo(period.microsecond),
  );

  /// Floors `this` to the nearest given `period`.
  DateTime floor(Period period) => _from(
    year.floorTo(period.year),
    month.floorTo(period.month),
    day.floorTo(period.day),
    hour.floorTo(period.hour),
    minute.floorTo(period.minute),
    second.floorTo(period.second),
    millisecond.floorTo(period.millisecond),
    microsecond.floorTo(period.microsecond),
  );


  DateTime _from(int year, int month, int day, int hour, int minute, int second, int millisecond, int microsecond) => isUtc
      ? GlobalDateTime(year, month, day, hour, minute, second, millisecond, microsecond)
      : LocalDateTime(year, month, day, hour, minute, second, millisecond, microsecond);

}


/// A `Chronological` contains methods to determine the chronological order of
/// multiple instances. Comparisons are independent of whether the time is in UTC
/// or local timezone.
mixin Chronological<T extends DateTime> on DateTime {

  bool operator <(T other) => isBefore(other);

  bool operator <=(T other) => compareTo(other) <= 0;

  @override
  bool operator ==(Object other) => other is DateTime && isAtSameMomentAs(other);

  bool operator >=(T other) => compareTo(other) >= 0;

  bool operator >(T other) => isAfter(other);

}


/// A date-time in the UTC timezone.
class GlobalDateTime extends DateTime with Chronological<GlobalDateTime> {

  /// Creates a `GlobalDateTime` with the given date and time.
  GlobalDateTime(int year, [
    int month = 1,
    int day = 1,
    int hour = 0,
    int minute = 0,
    int second = 0,
    int millisecond = 0,
    int microsecond = 0,
  ]): super.utc(year, month, day, hour, minute, second, millisecond, microsecond);

  /// Creates a `GlobalDateTime` with the given milliseconds since epoch.
  GlobalDateTime.epoch({@required int milliseconds}) : super.fromMillisecondsSinceEpoch(milliseconds, isUtc: true);

  /// Creates a `GlobalDateTime` that represents the current date and time.
  GlobalDateTime.now(): this._of(DateTime.now().toUtc());

  GlobalDateTime._of(DateTime date): this(date.year, date.month, date.day, date.hour, date.minute, date.second, date.millisecond, date.microsecond);


  /// Rounds `this` to the nearest given `period`.
  GlobalDateTime round(Period period) => GlobalDateTime(
    year.roundTo(period.year),
    month.roundTo(period.month),
    day.roundTo(period.day),
    hour.roundTo(period.hour),
    minute.roundTo(period.minute),
    second.roundTo(period.second),
    millisecond.roundTo(period.millisecond),
    microsecond.roundTo(period.microsecond),
  );

  /// Ceils `this` to the nearest given `period`.
  GlobalDateTime ceil(Period period) => GlobalDateTime(
    year.ceilTo(period.year),
    month.ceilTo(period.month),
    day.ceilTo(period.day),
    hour.ceilTo(period.hour),
    minute.ceilTo(period.minute),
    second.ceilTo(period.second),
    millisecond.ceilTo(period.millisecond),
    microsecond.ceilTo(period.microsecond),
  );

  /// Floors `this` to the nearest given `period`.
  GlobalDateTime floor(Period period) => GlobalDateTime(
    year.floorTo(period.year),
    month.floorTo(period.month),
    day.floorTo(period.day),
    hour.floorTo(period.hour),
    minute.floorTo(period.minute),
    second.floorTo(period.second),
    millisecond.floorTo(period.millisecond),
    microsecond.floorTo(period.microsecond),
  );


  /// Returns only the date.
  GlobalDateTime get date => GlobalDateTime(year, month, day);

}


/// A date-time in the local timezone.
class LocalDateTime extends DateTime with Chronological<LocalDateTime> {

  /// Creates a `LocalDateTime` with the given date and time.
  LocalDateTime(int year, [
      int month = 1,
      int day = 1,
      int hour = 0,
      int minute = 0,
      int second = 0,
      int millisecond = 0,
      int microsecond = 0,
  ]): super(year, month, day, hour, minute, second, millisecond, microsecond);

  /// Creates a `LocalDateTime` with the given milliseconds since epoch.
  LocalDateTime.epoch({@required int milliseconds}) : super.fromMillisecondsSinceEpoch(milliseconds);

  /// Creates a `LocalDateTime` that represents the current date and time.
  LocalDateTime.now(): super.now();

  LocalDateTime._of(DateTime date): this(date.year, date.month, date.day, date.hour, date.minute, date.second, date.millisecond, date.microsecond);


  /// Rounds `this` to the nearest given `period`.
  LocalDateTime round(Period period) => LocalDateTime(
    year.roundTo(period.year),
    month.roundTo(period.month),
    day.roundTo(period.day),
    hour.roundTo(period.hour),
    minute.roundTo(period.minute),
    second.roundTo(period.second),
    millisecond.roundTo(period.millisecond),
    microsecond.roundTo(period.microsecond),
  );

  /// Ceils `this` to the nearest given `period`.
  LocalDateTime ceil(Period period) => LocalDateTime(
    year.ceilTo(period.year),
    month.ceilTo(period.month),
    day.ceilTo(period.day),
    hour.ceilTo(period.hour),
    minute.ceilTo(period.minute),
    second.ceilTo(period.second),
    millisecond.ceilTo(period.millisecond),
    microsecond.ceilTo(period.microsecond),
  );

  /// Floors `this` to the nearest given `period`.
  LocalDateTime floor(Period period) => LocalDateTime(
    year.floorTo(period.year),
    month.floorTo(period.month),
    day.floorTo(period.day),
    hour.floorTo(period.hour),
    minute.floorTo(period.minute),
    second.floorTo(period.second),
    millisecond.floorTo(period.millisecond),
    microsecond.floorTo(period.microsecond),
  );


  /// Returns only the date.
  LocalDateTime get date => LocalDateTime(year, month, day);

}
