import 'package:sane/core.dart';

import 'time.dart';


/// Additional operations for a span of `DateTime`s.
extension DateTimeSpan on Span<DateTime> {

  /// Returns the duration between the [start] and [end].
  Duration get duration => end.difference(start);

}


/// Additional operations for a span of `Time`s.
extension TimeSpan on Span<Time> {

  /// Returns the duration between the [start] and [end].
  Duration get duration => end.difference(start);

}


/// Additional operations for a span of numbers.
extension NumericalSpan<T extends num> on Span<T> {

  /// Returns the duration between the [start] and [end].
  T get duration => end - start;

}


/// A length of time with a start and end.
class Span<T extends Comparable<dynamic>> implements Comparable<Span<T>> {

  /// The default [weight].
  static const int defaultWeight = 0;

  /// The start.
  final T start;
  /// The end.
  final T end;
  ///The weight used to determine the order of two spans that contain the same start and end.
  final int weight;
  int _hash;


  /// Creates a span with the given [start], [end] and [weight].
  /// If [end] is not given, the [start] is used also as the [end].
  /// If [weight] is not given, the [defaultWeight] is used.
  Span(this.start, [T end, this.weight = defaultWeight]): end = end ?? start;


  /// Returns `true` if `this` overlaps with `other` and `false` otherwise.
  bool overlaps(Span<T> other) => !(start.compareTo(other.start) <= 0 && end.compareTo(other.start) <= 0)
                               && !(start.compareTo(other.end) >= 0 && end.compareTo(other.end) >= 0);


  @override
  int compareTo(Span<T> other) {
    if (this == other) {
      return 0;
    }

    var comparison = start.compareTo(other.start);
    if (comparison != 0) {
      return comparison;
    }

    comparison = end.compareTo(other.end);
    if (comparison != 0) {
      return comparison;
    }

    return weight.compareTo(other.weight);
  }

  @override
  bool operator ==(Object other) => other is Span<T> && start == other.start && end == other.end && weight == other.weight;

  @override
  int get hashCode => _hash ??= hash([start, end, weight]);

  @override
  String toString() => 'Span{weight: $weight, duration: $start to $end}';

}



