import 'package:test/test.dart';

import 'package:sane/time.dart';


final DateTime local = DateTime(2019, 04, 13, 04, 31, 39, 364, 745);
final DateTime utc = DateTime.utc(2019, 04, 13, 04, 31, 39, 364, 745);


void main() {

  group('DateTimeConversion', () {

    group('toGlobal', () {

      test('local date', () {
        final global = local.toGlobal();

        expect(global.isUtc, isTrue);
        expect(global, local.toUtc());
      });

      test('utc date', () {
        final global = utc.toGlobal();

        expect(global.isUtc, isTrue);
        expect(global, utc);
      });

    });

    group('toLocalized', () {

      test('local date', () {
        final localized = local.toLocalized();

        expect(localized.isUtc, isFalse);
        expect(localized, local);
      });

      test('utc date', () {
        final localized = utc.toLocalized();

        expect(localized.isUtc, isFalse);
        expect(localized, utc.toLocal());
      });

    });

    group('date', () {

      test('local date', () {
        final date = local.date;

        expect(date.isUtc, isFalse);
        expect(date, isA<LocalDateTime>());
        expect(date, DateTime(2019, 04, 13));
      });

      test('utc date', () {
        final date = utc.date;

        expect(date.isUtc, isTrue);
        expect(date, isA<GlobalDateTime>());
        expect(date, DateTime.utc(2019, 04, 13));
      });

    });

  });

  group('RoundableDateTime', () {

    test('round', () {
      final rounded = local.round(const Period(
        year: 9,
        month: 7,
        day: 5,
        hour: 3,
        minute: 5,
        second: 7,
        millisecond: 9,
        microsecond: 11,
      ));

      expect(rounded, DateTime(2016, 7, 15, 03, 30, 42, 360, 748));
      expect(rounded, isA<LocalDateTime>());
    });

    test('ceil', () {
      final ceiled = utc.ceil(const Period(
        year: 9,
        month: 7,
        day: 5,
        hour: 3,
        minute: 5,
        second: 7,
        millisecond: 9,
        microsecond: 11,
      ));

      expect(ceiled, DateTime.utc(2025, 7, 15, 06, 35, 42, 369, 748));
      expect(ceiled, isA<GlobalDateTime>());
    });

    test('floor', () {
      final floored = utc.floor(const Period(
        year: 9,
        month: 7,
        day: 5,
        hour: 3,
        minute: 5,
        second: 7,
        millisecond: 9,
        microsecond: 11,
      ));

      expect(floored, DateTime.utc(2016, 0, 10, 03, 30, 35, 360, 737));
      expect(floored, isA<GlobalDateTime>());
    });

  });

  group('Chronological', () {

    final before = GlobalDateTime(2000);
    final after = GlobalDateTime(3000);

    test('<', () {
      expect(before < after, isTrue);
      expect(before < before, isFalse);
      expect(after < before, isFalse);
    });

    test('<=', () {
      expect(before <= after, isTrue);
      expect(before <= before, isTrue);
      expect(after <= before, isFalse);
    });

    group('==', () {

      // ignore: unrelated_type_equality_checks
      test('different type', () => expect(GlobalDateTime(1999) == 3, isFalse));

      test('supertype', () {
        expect(GlobalDateTime(1999), DateTime.utc(1999));
        expect(DateTime.utc(1999), GlobalDateTime(1999));

        expect(GlobalDateTime(1999), isNot(DateTime(1999)));
        expect(DateTime(1943), isNot(GlobalDateTime(1999)));
      });

      test('not same', () => expect(GlobalDateTime(1999), isNot(GlobalDateTime(1998))));

    });

    test('>=', () {
      expect(before >= after, isFalse);
      expect(before >= before, isTrue);
      expect(after >= before, isTrue);
    });

    test('>', () {
      expect(before > after, isFalse);
      expect(before > before, isFalse);
      expect(after > before, isTrue);
    });

  });

  group('GlobalDateTime', () {

    test('constructors are UTC-0', () {
      expect(GlobalDateTime(1).isUtc, isTrue);
      expect(GlobalDateTime.epoch(milliseconds: 5).isUtc, isTrue);
      expect(GlobalDateTime.now().isUtc, isTrue);
    });


    final global = GlobalDateTime(2019, 04, 13, 04, 31, 39, 364, 745);

    test('round', () {
      final rounded = global.round(const Period(
        year: 9,
        month: 7,
        day: 5,
        hour: 3,
        minute: 5,
        second: 7,
        millisecond: 9,
        microsecond: 11,
      ));

      expect(rounded, GlobalDateTime(2016, 7, 15, 03, 30, 42, 360, 748));
      expect(rounded, isA<GlobalDateTime>());
    });

    test('ceil', () {
      final ceiled = global.ceil(const Period(
        year: 9,
        month: 7,
        day: 5,
        hour: 3,
        minute: 5,
        second: 7,
        millisecond: 9,
        microsecond: 11,
      ));

      expect(ceiled, GlobalDateTime(2025, 7, 15, 06, 35, 42, 369, 748));
      expect(ceiled, isA<GlobalDateTime>());
    });

    test('floor', () {
      final floored = global.floor(const Period(
        year: 9,
        month: 7,
        day: 5,
        hour: 3,
        minute: 5,
        second: 7,
        millisecond: 9,
        microsecond: 11,
      ));

      expect(floored, GlobalDateTime(2016, 0, 10, 03, 30, 35, 360, 737));
      expect(floored, isA<GlobalDateTime>());
    });

    test('date', () {
      final date = global.date;

      expect(date.isUtc, isTrue);
      expect(date, isA<GlobalDateTime>());
      expect(date, GlobalDateTime(2019, 04, 13));
    });

  });

  group('LocalDateTime', () {

    test('constructors are local', () {
      expect(LocalDateTime(1).isUtc, isFalse);
      expect(LocalDateTime.epoch(milliseconds: 5).isUtc, isFalse);
      expect(LocalDateTime.now().isUtc, isFalse);
    });


    final localized = LocalDateTime(2019, 04, 13, 04, 31, 39, 364, 745);

    test('round', () {
      final rounded = localized.round(const Period(
        year: 9,
        month: 7,
        day: 5,
        hour: 3,
        minute: 5,
        second: 7,
        millisecond: 9,
        microsecond: 11,
      ));

      expect(rounded, LocalDateTime(2016, 7, 15, 03, 30, 42, 360, 748));
      expect(rounded, isA<LocalDateTime>());
    });

    test('ceil', () {
      final ceiled = localized.ceil(const Period(
        year: 9,
        month: 7,
        day: 5,
        hour: 3,
        minute: 5,
        second: 7,
        millisecond: 9,
        microsecond: 11,
      ));

      expect(ceiled, LocalDateTime(2025, 7, 15, 06, 35, 42, 369, 748));
      expect(ceiled, isA<LocalDateTime>());
    });

    test('floor', () {
      final floored = localized.floor(const Period(
        year: 9,
        month: 7,
        day: 5,
        hour: 3,
        minute: 5,
        second: 7,
        millisecond: 9,
        microsecond: 11,
      ));

      expect(floored, LocalDateTime(2016, 0, 10, 03, 30, 35, 360, 737));
      expect(floored, isA<LocalDateTime>());
    });

    test('date', () {
      final date = localized.date;

      expect(date.isUtc, isFalse);
      expect(date, isA<LocalDateTime>());
      expect(date, LocalDateTime(2019, 04, 13));
    });

  });

}