import 'package:test/test.dart';

import 'package:sane/time.dart';


void main() {

  group('TimeConversion', () {

    test('time', () => expect(DateTime.utc(0, 0, 0, 04, 10, 34, 358, 594).time, const Time(4, 10, 34, 358, 594)));

  });

  group('Time', () {

    group('enumerations', () {

      test('min', () => expect(const Time(0, 0, 0, 0, 0), Time.min));

      test('noon', () => expect(const Time(12, 0, 0, 0, 0), Time.noon));

      test('max', () => expect(const Time(23, 59, 59, 999, 999), Time.max));

    });

    group('assertions', () {

      group('hour', () {

        test('more than 24', () => expect(() => Time(24), throwsA(predicate<AssertionError>((e) => e.message == 'Invalid hour: 24, hour must be between 0 and 23'))));

        test('less than 0', () => expect(() => Time(-1), throwsA(predicate<AssertionError>((e) => e.message == 'Invalid hour: -1, hour must be between 0 and 23'))));

      });

      group('minute', () {

        test('more than 60', () => expect(() => Time(0, 60), throwsA(predicate<AssertionError>((e) => e.message == 'Invalid minute: 60, minute must be between 0 and 59'))));

        test('less than 0', () => expect(() => Time(0, -1), throwsA(predicate<AssertionError>((e) => e.message == 'Invalid minute: -1, minute must be between 0 and 59'))));

      });

      group('second', () {

        test('more than 60', () => expect(() => Time(0, 0, 60), throwsA(predicate<AssertionError>((e) => e.message == 'Invalid second: 60, second must be between 0 and 59'))));

        test('less than 0', () => expect(() => Time(0, 0, -1), throwsA(predicate<AssertionError>((e) => e.message == 'Invalid second: -1, second must be between 0 and 59'))));

      });

      group('millisecond', () {

        test('more than 1000', () => expect(() => Time(0, 0, 0, 1000), throwsA(predicate<AssertionError>((e) => e.message == 'Invalid millisecond: 1000, millisecond must be between 0 and 999'))));

        test('less than 0', () => expect(() => Time(0, 0, 0, -1), throwsA(predicate<AssertionError>((e) => e.message == 'Invalid millisecond: -1, millisecond must be between 0 and 999'))));

      });

      group('microsecond', () {

        test('more than 1000', () => expect(() => Time(0, 0, 0, 0, 1000), throwsA(predicate<AssertionError>((e) => e.message == 'Invalid microsecond: 1000, microsecond must be between 0 and 999'))));

        test('less than 0', () => expect(() => Time(0, 0, 0, 0, -1), throwsA(predicate<AssertionError>((e) => e.message == 'Invalid microsecond: -1, microsecond must be between 0 and 999'))));

      });

    });

    const before = Time(0, 1, 2, 3, 4);
    const after = Time(1, 3, 5, 7, 9);
    const duration = Duration(hours: 1, minutes: 2, seconds: 3, milliseconds: 4, microseconds: 5);

    group('addition and subtraction', () {

      group('invalid resultant time', () {

        test('exceeds maximum', () => expect(
            () => Time.max + const Duration(microseconds: 1),
            throwsA(predicate<RangeError>((e) => e.message == 'Invalid resultant time: 86400000000, time must be between 0 and 23:59:999:999'))
        ));

        test('exceeds minimum', () => expect(
            () => Time.min - const Duration(microseconds: 1),
            throwsA(predicate<RangeError>((e) => e.message == 'Invalid resultant time: -1, time must be between 0 and 23:59:999:999'))
        ));

      });

      test('addition', () => expect(before + duration, after));

      test('subtraction', () => expect(after - duration, before));

    });

    test('difference', () => expect(after.difference(before), duration));

    group('comparison', () {

      test('<', () {
        expect(Time.min < Time.max, isTrue);
        expect(Time.min < Time.min, isFalse);
        expect(Time.max < Time.min, isFalse);
      });

      test('<=', () {
        expect(Time.min <= Time.max, isTrue);
        expect(Time.min <= Time.min, isTrue);
        expect(Time.max <= Time.min, isFalse);
      });

      group('==', () {

        test('same type', () {
          expect(Time.min == Time.max, isFalse);
          expect(Time.min == Time.min, isTrue);
          expect(Time.min == const Time(0), isTrue);
        });

        // ignore: unrelated_type_equality_checks
        test('different type', () => expect(Time.min == const Duration(hours: 1), isFalse));

      });



      test('>=', () {
        expect(Time.min >= Time.max, isFalse);
        expect(Time.min >= Time.min, isTrue);
        expect(Time.max >= Time.min, isTrue);
      });

      test('>', () {
        expect(Time.min > Time.max, isFalse);
        expect(Time.min > Time.min, isFalse);
        expect(Time.max > Time.min, isTrue);
      });

    });

    test('hashCode', () => expect(Time.max.hashCode, maxMicroseconds));

    test('inMicroseconds', () => expect(Time.max.inMicroseconds, 86399999999));

    test('inMilliseconds', () => expect(Time.max.inMilliseconds, 86399999));

    test('toString', () {
      expect(Time.min.toString(), '00:00:00.000000');
      expect(Time.max.toString(), '23:59:59.999999');
      expect(const Time(0, 0, 0, 0, 1).toString(), '00:00:00.000001');
    });

  });



}